/* Put in the public domain 2000 by Sam Trenholme */
/* Kiwi Main:  The part of the Kiwi rouines with main() in it and the
   relevent command line parsers */

#include "JessStr/JessStr.h"
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <time.h>

/* Function prototypes */
void gencookie();
void decode();

/* Softerror: Exit with an error code allowing the MTA to try again */
int softerror(char *errormsg) {
    fprintf(stderr,"Soft error: %s\n",errormsg);
    exit(111); /* Exit code 111: Try again later */
    }

/* Harderror: Exit with an error code NOT allowing the MTA to try again */
/* As a general rule, use softerror when manipulations on non-user-controlled
   data go haywire, and use harderror when manipulations of user-controlled
   data go haywire */
int harderror(char *errormsg) {
    fprintf(stderr,"Fatal error: %s\n",errormsg);
    exit(100); /* Exit code 100: Return to sender */
    }

int main(int argc, char **argv) {
    js_string *arg, *tstr, *tstr2, *kiwirc;
    char badstr[256];
    char *envget; 
    int shift = 0;
    int tnum = 0;
    int offset = 0;

    /* Initialize js_string objects */
    if((arg = js_create(256,1)) == 0) 
        softerror("Unable to allocate arg.  Try later.");
    js_set_encode(arg,JS_US_ASCII);
    if((tstr = js_create(256,1)) == 0) 
        softerror("Unable to allocate tstr.  Try later.");
    js_set_encode(tstr,JS_US_ASCII);
    if((tstr2 = js_create(256,1)) == 0) 
        softerror("Unable to allocate tstr2.  Try later.");
    js_set_encode(tstr2,JS_US_ASCII);
    if((kiwirc = js_create(256,1)) == 0) 
        softerror("Unable to allocate kiwirc.  Try later.");
    js_set_encode(kiwirc,JS_US_ASCII);
    
    /* Regardless of anything else, we do need to read the user's .kiwirc
       file */
    if(find_kiwirc(kiwirc) == JS_ERROR)
        softerror("Error trying to locate .kiwirc file.");
    /* To do: Make sure the .kiwirc file has sane perms (600 or 700) */
    if(read_kiwirc(kiwirc,tstr,&tnum) == JS_ERROR)
        softerror("Error in reading .kiwirc file");
    if(tnum != 0) { /* If there was a syntax error in the .kiwirc file */
        if(tnum < -9000 || tnum > 9000)
	    softerror("Error number is out of bounds reading .kiwirc");
        sprintf(badstr,"%d",tnum); /* I really need a js_itoa function */
        js_qstr2js(tstr2,"Syntax error in .kiwirc file: ");
        js_append(tstr,tstr2); /* Add the error string */
        js_qstr2js(tstr,", line ");
        js_append(tstr,tstr2);
        js_qstr2js(tstr,badstr); /* Add the line number the error was on */
        js_append(tstr,tstr2);
        js_js2str(tstr2,badstr,100); /* Make it a NULL-terminated string */
        softerror(badstr);
        } 

    /* If argc is less than one (no arguments except argv[0]) we're haywire */
    if(argc < 1) 
        harderror("Argc is less than one!");
       
    /* Look at argv[0] and act accordingly */
    if(argv[0] != 0) {
        if(js_qstr2js(arg,argv[0]) == JS_ERROR) 
            harderror("Can not parse argv[0].");
        }
    else 
        harderror("argv[0] is a NULL pointer!");

    /* Now then, let's see how we should run Kiwi */

    /* Ignore anything leading up to a / */
    offset = 0;
    do {
        if(js_qstr2js(tstr,"/") == JS_ERROR)
            softerror("Can not give tstr a value, strange.");

        tnum = js_match_offset(tstr,arg,offset);
        if(tnum == JS_ERROR)
            harderror("Unable to run match on argv[0].");
    
        if(tnum != -2)
            offset = tnum + 1;

        } while(tnum != -2);

    /* See if the first letter of the command is a "c" */

    if(js_qstr2js(tstr,"c") == JS_ERROR) 
        softerror("Can not give tstr a value, strange.");

    tnum = js_match(tstr,arg);

    if(tnum == JS_ERROR)
        harderror("Unable to run match on argv[0].");
    if(tnum == offset) /* Command-line cookie generation */ {
        envget = (char *)getenv("REMOTE_ADDR");
        if(argc == 1) /* No arguments (execpt argv[0]) */
            gencookie(envget); /* This will be NULL if REMOTE_ADDR not set */	
        else if(argc == 2) { /* One argument */
            gencookie(argv[1]);
	    printf("\n");
	    }
	exit(0);
        } 

    if(js_qstr2js(tstr,"d") == JS_ERROR)
        softerror("Can not give tstr a value");

    tnum = js_match(tstr,arg);
    if(tnum == JS_ERROR)
        harderror("Unable to run match on argv[0]");
    if(tnum == offset) /* Decode cookie */ {
        if(argc == 1)
	    decode(0);
        else if(argc == 2) 
	    decode(argv[1]);
        exit(0);
	}

    harderror("Invalid argv[0] value.");

    }

void decode(char *crypto) {
    js_string *cookie, *key, *tstr;
    u_int32_t data;
    int counter;

    if((tstr = js_create(256,1)) == 0)
        softerror("Can not allocate temp string in gencookie");
    js_set_encode(tstr,JS_US_ASCII);
    if((cookie = js_create(256,1)) == 0)
        softerror("Can not allocate temp string in gencookie");
    js_set_encode(cookie,JS_US_ASCII);
    if((key = js_create(256,1)) == 0)
        softerror("Can not allocate key string in gencookie");
    js_set_encode(key,JS_US_ASCII);

    /* Get and set up the encryption key */
    js_qstr2js(tstr,"kiwi_key");
    read_kvar(tstr,key);
    kjs_key_setup(key);

    if(crypto == 0) { /* Get cookie to encrypt from user input */
        js_getline_stdin(cookie);
        if(js_length(cookie) != 8) /* Remember the newline! */
            harderror("Encrypted cookie must be seven characters long");
        }
    else { /* Get it from crypto argument */
        js_qstr2js(cookie,crypto);
        if(js_length(cookie) != 7)
            harderror("Encrypted cookie must be seven characters long");
        }

    data = kjs_parse_cookie(cookie);
    kjs_grokdata(data,tstr);
    js_show_stdout(tstr);
    printf("\n");
    }
 
/* Generate a Kiwi cookies, and output them to the standard output.
   input: the ip/5-letter code we wish to encrypt (NULL if timestamp crypted)
   output: none (outputs information to stdout)
   global vars used: But only in kead_kvar that this calls
*/
void gencookie(char *tocrypt) {
    /* Set up the key for encrypting data */
    js_string *tstr, *key;
    u_int32_t data;
    time_t time_s;

    if((tstr = js_create(256,1)) == 0)
        softerror("Can not allocate temp string in gencookie");
    js_set_encode(tstr,JS_US_ASCII);
    if((key = js_create(256,1)) == 0)
        softerror("Can not allocate key string in gencookie");
    js_set_encode(key,JS_US_ASCII);

    /* Get and set up the encryption key */
    js_qstr2js(tstr,"kiwi_key");
    read_kvar(tstr,key);
    kjs_key_setup(key);

    if(tocrypt == 0) { /* Timestamp encryption */ 
        /* Get the time stamp and make cookies for short, mid, and long 
           timeout based on it */
        time_s = time(0);

        /* Short timeout */
        data = 0x80000000;
        data |= (time_s & 0xFFFFFFFF) >> 4; 
        kjs_make_cookie(data,tstr);
        printf("%s","Cookie with short timeout: ");
        js_show_stdout(tstr);
        printf("\n");

        /* Mid timeout */
        data = 0x90000000;
        data |= (time_s & 0xFFFFFFFF) >> 4; 
        kjs_make_cookie(data,tstr);
        printf("%s","Cookie with medium timeout (Usenet cookie): ");
        js_show_stdout(tstr);
        printf("\n");
 
        /* Long timeout */
        data = 0xA0000000;
        data |= (time_s & 0xFFFFFFFF) >> 4; 
        kjs_make_cookie(data,tstr);
        printf("%s","Cookie with long timeout: ");
        js_show_stdout(tstr);
        printf("\n");

        return;
        }
    else { /* Argument was given to tocrypt */

        if(js_qstr2js(tstr,tocrypt) == JS_ERROR)
	    harderror("Error converting REMOTE_ADDR/argv[1]");

        if(js_length(tstr) > 6) /* IP to crypt */ {
            int place = 0; /* Place to look for an octet */
            u_int8_t ipo[4];
            int counter;
	    js_string *dotq;
	    
	    if((dotq = js_create(256,1)) == 0)
	        softerror("Can not allocate memory for dotq");
           
	    js_set_encode(dotq,JS_US_ASCII);
	    js_qstr2js(dotq,".");

            for(counter = 0;counter < 4;counter++) {
                if(place == -2) 
		    harderror("IP must be in dotted octet form like 10.1.1.1");
                else {
                    if(place != 0)
                        place++;
	            ipo[counter] = js_atoi(tstr,place);
                    }
		place = js_match_offset(dotq,tstr,place);
		if(place == JS_ERROR)
		    harderror("Problem matching on IP string");
                }
            
	    data = (ipo[0] << 24) | (ipo[1] << 16) | (ipo[2] << 8) | ipo[3];
	    data >>= 4;
	    data |= 0xC0000000;
	    kjs_make_cookie(data,tstr);
	    js_show_stdout(tstr); 
	    return;
	    }
        else { /* 5-letter message to crypt */
	    int val,counter;
            data = 0;

            if(js_qstr2js(tstr,tocrypt) == JS_ERROR)
	        harderror("Unable to js-ize 5-letter message!");

            if(js_length(tstr) != 5)
	        harderror("Message must be exactly 5 letters long.");

            for(counter=0;counter<5;counter++) {
	        data <<= 5;

		val = js_val(tstr,counter);
                if(val == JS_ERROR)
                    harderror("Problem getting value of message char");
                
		val = us_ascii_unmap(val);
                if(val == JS_ERROR || val < 0 || val >= 26)
                    harderror("Please use only lower case letters in message");

                data |= val & 31;
                }

	    data |= 0x7C000000;
            kjs_make_cookie(data,tstr);
	    js_show_stdout(tstr);
	    return;
            }
        } 

    /* Something wrong if we get here */
    harderror("Invalid data given to clicrypt!");
    }

